open Js      ;;
open JSOO    ;;
open Base    ;;
open Process ;;

(*****************************
 *           UTILS           *
 *****************************)


let examples =  [ "simple.wc"
                ; "first.wc"
                ; "arith.wc"
                ; "if.wc"
                ; "rec.wc"
                ; "echo.wc"
                ; "higher.wc"
                ]


(** Tells which method to run *)
type process = TeX
             | Shell
             | PAst
             | PCore
             | PAlpha
             | PMoved
             | PEta
             | PSuper
             | PPseudo

let actions = [ (TeX     , "TeX"                        )
              ; (Shell   , "Shell"                      )
              ; (PAst    , "Abstract-Syntax Tree"       )
              ; (PCore   , "Core intermediate language" )
(*              ; (PAlpha  , "Alpha renamed Core"         )
              ; (PMoved  , "Moved variable Core"        )
              ; (PEta    , "Eta reduced Core"           ) *)
              ; (PSuper  , "Super Combinators"          )
              ; (PPseudo , "Pseudo code"                )
              ]

(*****************************
 *         ELEMENTS          *
 *****************************)


let textarea ?style ?onkeyup ?(attrs = []) children =
   let m = Html.create "textarea" ~attrs:attrs () in
      Html.set_attr_opt m "style" style ;
      Html.register_event_opt m "onkeyup" (match onkeyup with
                                            | None -> None
                                            | Some f -> Some (fun () -> f m)
                                          ) ;
      List.iter (Node.append m) children ;
      m


let h3 ?attrs s =
 let m = Html.create "h3" ?attrs:attrs ()
 in Node.append m (Node.text s) ;
    m

let listnodes s l =
 Html.div ((h3 s)
          :: List.concat (List.map (fun x -> [Html.string " -> " ; Base.fst x]) l)
          )



(*****************************
 *   CREATING THE DOCUMENT   *
 *****************************)

let camlzone    = get_element_by_id "camlzone"   ;;
let status      = Html.div ~attrs:[("id","status")] [] ;;

let inputcode   = textarea ~attrs:[ ("id"      ,"inputcode" )
                                  ; ("rows"    ,"30"        )
                                  ; ("cols"    ,"50"        )
                                  ; ("wrap"    , "off"      )
                                  ] [] ;;

inputcode  >>> set "value" (string "(* !!!!!! Type here your code !!!!!\n   For example:\n *)\n\n let f = fun x -> (x 7) in f\n" )
;;

let outputcode  = textarea  ~attrs:[ ("id"      ,"outputcode")
                                   ; ("rows"    ,"30"        )
                                   ; ("cols"    ,"80"        )
                                   ; ("wrap"    , "off"      )
                                   ; ("readonly","yes"      )
                                   ] [] ;;
outputcode >>> set "value" (string " The result will appear here! " )
;;

let examplenodes = List.map (fun x     -> (Html.a [Node.text x] , x )) examples ;;
let actionnodes  = List.map (fun (x,y) -> (Html.a [Node.text y] , x )) actions  ;;


Node.append camlzone ( Html.div [
  listnodes "Choose a file to load (Or type our own!)"       examplenodes ;
  listnodes "Choose a compilation method"  actionnodes  ;
  status ;
  Html.br () ;
  inputcode ;
  outputcode
] ) ;;



(*****************************
 *         PAGE API          *
 *****************************)


let setStatus s = Node.replace_all status (Node.text s) ;;

let loadfile file =
 let s   = "examples/" ^ file                       in
 let _   = setStatus ("Loading file " ^ s ^ " ...") in
 let str = http_get s                               in
 let _   = setStatus ("Setting input value ...")    in
 let _   = inputcode >>> set "value" (string str)   in
 setStatus "Done!"


let process a =
 let str        = inputcode >>> get "value" >>> as_string in
 let _          = setStatus "Creating lexing ..."         in
 let lex        = Lexing.from_string str                  in
 let _          = setStatus "Parsing ast ..."             in
 let ast        = Parser.ast Lexer.lexer lex              in
 let texget   s = http_get ("tex/"   ^ s ^ ".tex")        in
 let shellget s = http_get ("shell/" ^ s         )        in
 let _          = setStatus "Compiling ...."              in
 let out        = (match a with
                     | TeX     -> processTeX    texget    ast
                     | Shell   -> processShell  shellget  ast
                     | PAst    -> processAst              ast
                     | PCore   -> processCore             ast
                     | PAlpha  -> processAlpha            ast
                     | PMoved  -> processMoved            ast
                     | PEta    -> processEta              ast
                     | PSuper  -> processSuper            ast
                     | PPseudo -> processPseudo           ast     
                  ) in
 let _          = setStatus "Outputing ..."                 in
 let _          = outputcode >>> set "value" (string out)
 in setStatus "Done!" ;;


(*****************************
 *    REGISTERING EVENTS     *
 *****************************)


let register_click_list f l =
  List.iter (fun (obj,v) -> Node.register_event obj "onclick" f v) l ;;

register_click_list loadfile examplenodes ;;
register_click_list process  actionnodes ;;

